
# Class.tcl
#
# ClassDesigner Class meta-object handling
#

package provide Class 0.1;

class Class {

  extends Element;

  import Output;
  import Indent;
  import Comment;
  import Console;
  import library;

  staticarray template;
  field container;
  field innerclasses;

  source [file join [library] TmplClas.tcl];

  method init {} {
    set container "";
    set innerclasses "";

    foreach operation [$self get_operations] {
      new Operation $operation $self;
    }
    foreach attribute [$self get_attributes] {
      new Attribute $attribute $self;
    }
    foreach assocend [$self get_assoc_ends] {
      new AssociationEnd $assocend;
    }
  }

  method setContainer {outerclass} {
    set container $outerclass;
  }

  method addInnerclass {innerclass} {
    lappend innerclasses $innerclass;
  }

  #
  # Returns the (recursive set of) interfaces this class implements
  #
  method interfaces {} {
    set interfaces {};
    set generalizations [$self get_generalizations];
    foreach generalization $generalizations {
      set superclass [$generalization get_superclass];
      if { ![$superclass is interface] } continue;
      set interfaces \
	  [concat $interfaces [concat $superclass [$superclass interfaces]]];
    }
    
    return $interfaces;
  }

  # 
  # Returns the (immediate) superclass of this class
  #
  method superclass {} {
    # look up the first real, i.e. non-interface superclass
    set generalizations [$self get_generalizations];
    foreach generalization $generalizations {
      set superclass [$generalization get_superclass]; 
      if { $superclass == "" } continue;
      if { [$superclass is interface] } continue;
      return "$superclass";
    }
  }

  #
  # Returns the (recursive set of) non-interface superclasses
  #
  method superclasses {} {
    set generalizations [$self get_generalizations];
    foreach generalization $generalizations {
      set superclass [$generalization get_superclass];
      if { [$superclass is interface] } continue;
      return [concat $superclass [$superclass superclasses]];
    }
  }

  #
  # Creates an virtual method which is mapped over elements of a list
  #
  staticmethod map {list code} {
    set result {};
    foreach self $list {
      lappend result [eval $code];
    }
    set result [join $result "\, "];
    return $result;
  }

  #
  # Verifies the integrity of the class
  #
  method violations {{mode verbose}} {
    set name [$self get_name];
    set num_errors 0;

    #
    # Verify that the container class (if any) of the class
    # has been properly identified
    #
    set outerclass [$self prop "innerclassof"];
    if { $outerclass != "" && $container == "" } {
      $self error "$name: unable to resolve the container class \
	  ( $outerclass )" $mode;
      incr num_errors;
    }

    #
    # Multiple inheritance disallowed
    #
    set superclasses {};
    set generalizations [$self get_generalizations];
    foreach generalization $generalizations {
      set superclass [$generalization get_superclass]; 
      if { [$superclass is interface] } continue;
      lappend superclasses "$superclass";
    }
    if { [llength $superclasses] > 1 } {
      $self error "$name: multiple superclasses \
	  ( [map $superclasses {$self get_name}] )" $mode;
      incr num_errors;
    }

    #
    # Verify integrity of attributes
    #
    set attributes [$self get_attributes];
    foreach attribute $attributes {
      incr num_errors [$attribute violations $mode];
    }

    #
    # A non-abstract class can't have abstract methods;
    # an abstract class should have at least one abstract method.
    #
    set absOps 0;
    set operations [$self get_operations];
    foreach operation $operations {
      incr num_errors [$operation violations $mode];
      if { [$operation is abstract] } {
	incr absOps 1;
	if { ! ([$self is abstract] || [$self is interface]) } {
	  $self error "$name: non-abstract class with\
	      an abstract method ( [$operation get_signature] )" $mode;
	  incr num_errors;
	}
      }
    }
    if { $absOps == 0 && [$self is abstract] } {
      $self warning "$name: abstract class with no abstract methods" $mode;
    }

    #
    # XXX Should check whether the class implements the abstract methods
    # XXX it promises to implement
    #
    
    #
    # Verify integrity of inner classes
    #
    foreach innerclass $innerclasses {
      Console message "Verifying inner class [$innerclass get_name]...";
      incr num_errors [$innerclass violations $mode];
    }

    return $num_errors;
  }

  #
  # Prints the class
  #
  method print {} {
    set name [$self get_name];

    #
    # Print description
    #
    if { $container == "" } {
      Comment print description "$name\n\n[$self get_description]";
    } else {
      Comment print description "$name (inner class)\n\n[$self get_description]";
    }

    #
    # Print beginning
    #
    set expander [new Expander];

    if { $container != "" } {
      $expander bind visibility "";
    } else {
      $expander bind visibility "[$self prop visibility]";
    }

    $expander bind abstract   "[$self prop abstract]";
    $expander bind final      "[$self prop final]";
    $expander bind name	      "$name";

    set superclass [$self superclass];
    if { $superclass != "" } {
      $expander bind superclass "[$superclass get_name]";
    } else {
      $expander bind superclass "Object";
    }
    $expander bind interfaces "[map [$self interfaces] {$self get_name}]";

    set stereotype [$self get_stereotype];
    if { $stereotype != "<unspecified>" && \
	[lsearch [array names template] $stereotype] > 0 } {
      Output print [$expander expand $template($stereotype)];
    } else {
      Output print [$expander expand $template(class)];
    }
 
    Indent more;

    #
    # Print attributes
    # 
    set attributes [$self get_attributes];
    foreach attr $attributes {
      $attr print;
    }

    #
    # Print associations
    #
    set ends [$self get_assoc_ends];
    foreach end $ends {
      $end print;
    }

    # 
    # Print associations defined in direct interfaces
    #
    set interfaces [$self interfaces];
    foreach interface $interfaces {
      set ends [$interface get_assoc_ends];
      foreach end $ends {
        $end print;
      }
    }

    #
    # Print operations
    #
    set operations [$self get_operations];
    foreach operation $operations {
      $operation print;
    }

    #
    # Print bound and constrained property registration
    #
    set boundprops {};
    set constrprops {};
    set attributes [$self get_attributes];
    foreach attr $attributes {
      set name [$attr get_name];
      set beanprop [$attr getProperty "beanprop"];
      switch $beanprop {
	bound { lappend boundprops $name }
	constrained { lappend constrprops $name }
      }
    }      
    $expander bind boundprops "$boundprops";
    $expander bind constrprops "$constrprops";
    if { [llength $boundprops] > 0 } {
      Output print [$expander expand $template(bean-bound)];
    }
    if { [llength $constrprops] > 0 } {
      Output print [$expander expand $template(bean-constrained)];
    }

    #
    # Print inner classes
    #
    foreach innerclass $innerclasses {
      Console message "Generating code for inner class [$innerclass get_name]...";
      $innerclass print;
    }
 
    #
    # Print end
    #
    Indent less;

    Output print [$expander expand $template(end)];

    delete $expander;
  }
}

