#
# Element.tcl
#
# Routines to handle common issues with elements (classes, fields etc)
#

package provide Element 0.1;

class Element {

  extends TclDecorator;

  import Console;

  method hasProperty {property} {
    set p1 [$self get_property "$property"];
    set p2 [$self get_property " $property "];
    return [expr {$p1 == "<unspecified>"} && {$p2 == "<unspecified>"} ? 0: 1]
  }

  method getProperty {property} {
    set p [$self get_property $property];
    if { $p == "<unspecified>" } {
      return "";
    } else {
      return "$p";
    }
  }

  #
  # Sanity checks omitted on purpose to keep the method compact,
  # so it is relied on that e.g. an Operation is never tested against
  # "is transient".
  #
  method is {aspect} {
    switch "$aspect" {
      abstract     { return [$self hasProperty "abstract"]; }
      static       {
	if { [$self is class] || [$self is interface] } {
	  return [$self hasProperty "static"];
	} else {
	  return [$self get_is_classifier_scope]; # bug in ClassDesigner's docs
	}
      } 
      final        { return [$self hasProperty "final"]; }
      native       { return [$self hasProperty "native"]; }
      synchronized { return [$self hasProperty "synchronized"]; }
      volatile     { return [$self hasProperty "volatile"]; }
      transient    { return [$self hasProperty "transient"]; }
      disabled     { return [$self hasProperty "disabled"]; }
      beanprop     { return [$self hasProperty "beanprop"]; }
      class        { return [$self is_kind_of "Class"]; }
      interface    { return [$self is_kind_of "Interface"]; }
    }
  }

  method prop {aspect} {
    switch "$aspect" {
      abstract    -
      static      -
      final       -
      native	  -
      synchronized -
      volatile    -
      transient   { 
	if { [$self is $aspect] } { 
 	  return "$aspect "; 
	} else {
	  return "";
	}
      }

      visibility {
	set vis "";
	if { [$self is_kind_of "Feature"] } {
	  if { [$self hasProperty "visibility"] } {
	    set vis [$self getProperty "visibility"];
	  }
	  if { $vis == "" } {
	    set vis [$self get_visibility];
	  }
	} else {
	  set vis [$self get_namespace_visibility];
	}
	if { $vis == "unspecified" } {
	  return "";
	}
	return "$vis ";
      }

      disabled {
	set disabled "";
	if { [$self hasProperty "disabled"] } {
	  set items "[split [$self getProperty "disabled"] ,]";
	  foreach item $items {
	    regexp {^ *([a-z]+) *$} $item all result;
	    if { $result != "" } {
	      lappend disabled $result;
	    }
	  } 
	}
	return "$disabled";
      }
      
      beanprop {
	if { ! [$self hasProperty "beanprop"] } {
	  return "";
	}
	switch [$self get_type] {
	  boolean { set type "boolean" }
	  Boolean { set type "boolean" }
	  default { set type "simple" }
	}
	set mod [$self getProperty "beanprop"];
	if { $mod != "" } {
	  append type "-";
	}
	return "$type$mod";
      }

      exceptions {
        if { [$self hasProperty "exceptions"] } {
          return "[$self getProperty "exceptions"] ";
        } else {
          return "";
        }
      }

      innerclassof {
	if { [$self hasProperty "innerclassof"] } {
	  return "[$self getProperty "innerclassof"]";
	} else {
	  return "";
	}
      }
    }
  }

  method plainType {type} {
    set result "";
    regsub {<.*>} $type "" result;

    return $result;
  }
  
  method parseType {type} {
    set result "";
    regsub {<} $type " /* " result;
    regsub {>} $result " */ " result;

    return $result;
  }

  method warning {msg {mode verbose}} {
    if { $mode == "verbose"} {
      Console warning "$msg";
    }
  }

  method error {msg {mode verbose}} {
    if { $mode == "verbose"} {
      Console error "$msg";
    }
  }
}
