#
# Interfac.tcl
#
# ClassDesigner Interface meta-object handling
#

package provide Interface 0.1;

class Interface {

  extends Element;

  import Output;
  import Indent;
  import Comment;
  import Console;
  import library;

  staticarray template;
  field container;

  source [file join [library] TmplInfc.tcl];

  method init {} {
    set container "";

    foreach operation [$self get_operations] {
      new Operation $operation $self;
    }
    foreach attribute [$self get_attributes] {
      new Attribute $attribute $self;
    }
    foreach assocend [$self get_assoc_ends] {
      new AssociationEnd $assocend;
    }
  }

  method setContainer {outerclass} {
    set container $outerclass;
  }

  #
  # Returns the (recursive set of) interfaces this interface extends
  #
  method interfaces {} {
    set interfaces {};
    set generalizations [$self get_generalizations];
    foreach generalization $generalizations {
      set superclass [$generalization get_superclass];
      if { ![$superclass is interface] } continue;
      set interfaces \
	  [concat $interfaces [concat $superclass [$superclass interfaces]]];
    }
    return $interfaces;
  }

  #
  # Creates a virtual method which is mapped over elements of a list
  #
  staticmethod map {list code} {
    set result {};
    foreach self $list {
      lappend result [eval $code];
    }
    set result [join $result "\, "];
    return $result;
  }

  #
  # Verifies the integrity of the interface
  #
  method violations {{mode verbose}} {
    set name [$self get_name];
    set num_errors 0;

    #
    # No superclasses allowed
    #
    set generalizations [$self get_generalizations];
    foreach generalization $generalizations {
      set superclass [$generalization get_superclass]; 
      if { $superclass == "" } continue;
      if { [$superclass is interface] } continue;
      $self error "$name: interface extending a class \
	  ( [$superclass get_name] )" $mode;
      incr num_errors;
    }

    #
    # Non-abstract methods (and non-static/non-final fields)
    # are not considered an integrity violation;
    # they are silently forced to "abstract" (to final and static).
    # See Operation and Attribute.
    #
    return $num_errors;
  }

  #
  # Prints the interface
  #
  method print {} {
    set name [$self get_name];

    #
    # Print description
    #
    if { $container == "" } {
      Comment print description "$name\n\n[$self get_description]";
    } else {
      Comment print description "$name (inner interface)\n\n[$self get_description]";
    }

    #
    # Print beginning
    #
    set expander [new Expander];

    if { $container != "" } {
      $expander bind visibility "";
    } else {
      $expander bind visibility "[$self prop visibility]";
    }

    $expander bind abstract   "[$self prop abstract]";
    $expander bind final      "[$self prop final]";
    $expander bind name	      "$name";
    $expander bind interfaces "[map [$self interfaces] {$self get_name}]";

    Output print [$expander expand $template(interface)];
 
    Indent more;

    #
    # Print attributes
    # 
    set attributes [$self get_attributes];
    foreach attr $attributes {
      $attr print;
    }

    #
    # Print operations
    #
    set operations [$self get_operations];
    foreach operation $operations {
      $operation print;
    }

    #
    # Print end
    #
    Indent less;

    Output print [$expander expand $template(end)];

    delete $expander;
  }
}

